import collections
import importlib


class PLL:
    def __init__(self, fs, cuda=False, numba=False):
        # Import Dynamic Modules
        self.load_modules(cuda, numba)

        # Variables to Self
        self.fs = fs
        self.freq = 0.0
        self.times = None
        self.f_cb = collections.deque(maxlen=128)
        self.algn = self.xp.arange(0, self.xp.pi*2, self.xp.pi/4)

    def load_modules(self, cuda, numba):
        self.cuda = cuda
        self.numba = numba

        if self.cuda:
            self.xp = importlib.import_module('cupy')
            self.np = importlib.import_module('numpy')
            self.ts = importlib.import_module('radio.tools.pll._core_cuda')
        elif not self.numba:
            self.xp = importlib.import_module('numpy')
            self.ts = importlib.import_module('radio.tools.pll._core')
            self.np = self.xp

    def wave(self, freq, phase, times):
        return self.xp.cos(2.0*self.xp.pi*freq*times+phase)

    def step(self, x):
        # Make Reference Phase
        if self.times is None:
            self.times = self.xp.arange(0, 1, 1/self.fs)[:len(x)]

        # Estimate Signal Frequency
        freq = self.ts.frequency(x, self.fs)
        self.f_cb.append(freq)
        self.freq = self.np.mean(self.f_cb)

        # Get Phase Compensation
        error_ls = self.ts.alignment(x, self.algn, self.freq, self.times)
        self.phi = self.algn[self.xp.argmax(error_ls)]

    def mult(self, mult=1):
        omega = ((self.phi*mult)+(self.xp.pi/mult))
        return self.wave(self.freq*mult, omega, self.times)



def lfilter(self, b, a, x, zi):
    if self.cuda:
        x = self.xp.asnumpy(x)

    x, zi = self.ss.lfilter(b, a, x, zi=zi)

    if self.cuda:
        x = self.xp.asarray(x)

    return (x, zi)


def filtfilt(self, b, a, x):
    if self.cuda:
        x = self.xp.asnumpy(x)

    x = self.ss.filtfilt(b, a, x)

    if self.cuda:
        x = self.xp.asarray(x)

    return x



import collections
import importlib

from radio.tools.pll import PLL
from radio.tools.helpers import lfilter, filtfilt


class WBFM:

    def __init__(self, tau, ifs, ofs, cuda=False, numba=False):
        # Import Dynamic Modules
        self.load_modules(cuda, numba)

        # Variables to Self
        self.tau = tau
        self.ifs = ifs
        self.ofs = ofs
        self.dec = int(self.ifs/self.ofs)

        # Check Parameters
        assert (self.ifs % self.ofs) == 0

        # Setup Pilot PLL
        self.pll = PLL(self.ifs, cuda=self.cuda)
        self.freq = self.pll.freq

        # Setup Filters
        x = self.np.exp(-1/(self.ofs * self.tau))
        pb, pa = self.ss.butter(2, [19e3-200, 19e3+200], btype='band', fs=self.ifs)
        mb, ma = self.ss.butter(8, 15e3, btype='low', fs=self.ifs)

        self.fi = {
            "db": [1-x], "da": [1, -x],
            "pb": pb, "pa": pa,
            "mb": mb, "ma": ma,
        }

        # Setup Filters Initial Conditions
        self.zi = {
            "mlpr": self.ss.lfilter_zi(mb, ma),
            "mlmr": self.ss.lfilter_zi(mb, ma),
            "dlpr": self.ss.lfilter_zi(self.fi["db"], self.fi["da"]),
            "dlmr": self.ss.lfilter_zi(self.fi["db"], self.fi["da"]),
        }

        # Setup continuity data
        self.co = {
            "dc": collections.deque(maxlen=32),
        }

    def load_modules(self, cuda, numba):
        self.cuda = cuda
        self.numba = numba

        if self.cuda:
            self.xs = importlib.import_module('cusignal')
            self.xp = importlib.import_module('cupy')
            self.np = importlib.import_module('numpy')
            self.ss = importlib.import_module('scipy.signal')
        else:
            self.xs = importlib.import_module('scipy.signal')
            self.xp = importlib.import_module('numpy')
            self.np = self.xp
            self.ss = self.xs

    def run(self, buff):
        b = self.xp.array(buff)
        b = self.xp.angle(b)
        b = self.xp.unwrap(b)
        b = self.xp.diff(b)
        b = self.xp.concatenate((b, self.xp.array([b[-1]])))
        b /= self.xp.pi

        # Normalize for DC
        dc = self.xp.mean(b)
        self.co['dc'].append(dc)
        b -= self.np.mean(self.co['dc'])

        # Synchronize PLL with Pilot
        P = filtfilt(self, self.fi['pb'], self.fi['pa'], b)
        self.pll.step(P)

        # Demod Left + Right (LPR)
        LPR, self.zi['mlpr'] = lfilter(self, self.fi['mb'], self.fi['ma'], b, zi=self.zi['mlpr'])
        LPR = self.xs.resample_poly(LPR, 1, self.dec, window='hamm')
        LPR, self.zi['dlpr'] = lfilter(self, self.fi['db'], self.fi['da'], LPR, zi=self.zi['dlpr'])

        # Demod Left - Right (LMR)
        LMR = (self.pll.mult(2) * b) * 1.0175
        LMR, self.zi['mlmr'] = lfilter(self, self.fi['mb'], self.fi['ma'], LMR, zi=self.zi['mlmr'])
        LMR = self.xs.resample_poly(LMR, 1, self.dec, window='hamm')
        LMR, self.zi['dlmr'] = lfilter(self, self.fi['db'], self.fi['da'], LMR, zi=self.zi['dlmr'])

        # Mix L+R and L-R to generate L and R
        L = LPR+LMR
        R = LPR-LMR

        # Ensure Bounds
        L = self.xp.clip(L, -0.99, 0.99)
        R = self.xp.clip(R, -0.99, 0.99)

        if self.cuda:
            L = self.xp.asnumpy(L)
            R = self.xp.asnumpy(R)

        return L, R



import collections
import importlib

from radio.tools.helpers import lfilter


class MFM:

    def __init__(self, tau, ifs, ofs, cuda=False, numba=False):
        # Import Dynamic Modules
        self.load_modules(cuda, numba)

        # Variables to Self
        self.tau = tau
        self.ifs = ifs
        self.ofs = ofs
        self.dec = int(self.ifs/self.ofs)

        # Check Parameters
        assert (self.ifs % self.ofs) == 0

        # Make De-emphasis Filter
        x = self.np.exp(-1/(self.ofs * self.tau))
        self.db = [1-x]
        self.da = [1, -x]
        self.zi = self.ss.lfilter_zi(self.db, self.da)

        # Setup continuity data
        self.co = {
            "dc": collections.deque(maxlen=32),
        }

    def load_modules(self, cuda, numba):
        self.cuda = cuda
        self.numba = numba

        if self.cuda:
            self.xs = importlib.import_module('cusignal')
            self.xp = importlib.import_module('cupy')
            self.np = importlib.import_module('numpy')
            self.ss = importlib.import_module('scipy.signal')
        else:
            self.xs = importlib.import_module('scipy.signal')
            self.xp = importlib.import_module('numpy')
            self.np = self.xp
            self.ss = self.xs

    def run(self, buff):
        b = self.xp.array(buff)
        b = self.xp.angle(b)
        b = self.xp.unwrap(b)
        b = self.xp.diff(b)
        b = self.xp.concatenate((b, self.xp.array([b[-1]])))
        b /= self.xp.pi

        # Normalize for DC
        dc = self.xp.mean(b)
        self.co['dc'].append(dc)
        b -= self.np.mean(self.co['dc'])

        # Demod Left + Right (LPR)
        LPR = self.xs.resample_poly(b, 1, self.dec, window='hamm')
        LPR, self.zi = lfilter(self, self.db, self.da, LPR, zi=self.zi)

        # Ensure Bounds
        LPR = self.xp.clip(LPR, -0.99, 0.99)

        if self.cuda:
            return self.xp.asnumpy(LPR)

        return LPR



import numpy as np
from typing import List


def frequency(x, fs) -> float:
    zeros = np.where(np.diff(np.signbit(x)))[0]
    return float(fs/np.mean(np.diff(zeros[5:-5]))/2)


def alignment(x, algn, freq, times) -> List[float]:
    err_ls = []
    for i in range(len(algn)):
        signal = x + np.sin(2.0*np.pi*freq*times+algn[i])
        err_ls.append(np.mean(np.abs(signal)))
    return np.asarray(err_ls)



import cupy as cp
from typing import List


def frequency(x, fs) -> float:
    zeros = cp.where(cp.diff(cp.signbit(x)))[0]
    return fs/cp.mean(cp.diff(zeros[5:-5]))/2


def alignment(x, algn, freq, times) -> List[float]:
    err_ls = []
    for i in range(len(algn)):
        signal = x + cp.sin(2.0*cp.pi*freq*times+algn[i])
        err_ls.append(cp.mean(cp.abs(signal)))
    return cp.asarray(err_ls)



import importlib


class Tuner:

    def __init__(self, bands, osize, cuda=False):
        # Import Dynamic Modules
        self.load_modules(cuda)

        # Variables to Self
        self.bands = bands

        # List Bands Boundaries
        los = self.xp.array([(b['freq'] - (b['bw']/2)) for b in self.bands])
        his = self.xp.array([(b['freq'] + (b['bw']/2)) for b in self.bands])

        # Get Bands Boundaries
        self.lof = self.xp.min(los)
        self.hif = self.xp.max(his)

        # Get Bandwidth and Center Frequency
        self.bw = float(self.hif-self.lof)
        self.bw += self.bands[0]['bw']-(self.bw % self.bands[0]['bw'])
        self.mdf = float((self.lof+self.hif)/2.0)

        # Get Size
        self.size = int(osize*(self.bw//self.bands[0]['bw']))

        # List Decimation Factor
        self.dfac = [int(self.size/(self.bw//b['bw'])) for b in self.bands]

        # List Frequency & FFT Offset
        self.foff = [b['freq'] - self.mdf for b in self.bands]
        self.toff = [-(self.size*f)/self.bw for f in self.foff]

    def load_modules(self, cuda):
        self.cuda = cuda

        if self.cuda:
            self.xs = importlib.import_module('cusignal')
            self.xp = importlib.import_module('cupy')
            self.np = importlib.import_module('numpy')
            self.ss = importlib.import_module('scipy.signal')
        else:
            self.xs = importlib.import_module('scipy.signal')
            self.xp = importlib.import_module('numpy')
            self.np = self.xp
            self.ss = self.xs

    def load(self, buff):
        self.b = self.xp.fft.fft(self.xp.array(buff))

    def run(self, id):
        a = self.xp.roll(self.b, int(self.toff[id]))
        a = self.xs.resample(a, self.dfac[id], window='hamm', domain="freq")
        return a



import importlib


class Decimator:

    def __init__(self, ifs, ofs, cuda=False):
        # Import Dynamic Modules
        self.load_modules(cuda)

        # Variables to Self
        self.out_size = -1
        self.dec = ifs/ofs

        print("[DECIMATOR] Factor: {}".format(self.dec))

    def load_modules(self, cuda):
        self.cuda = cuda

        if self.cuda:
            self.xs = importlib.import_module('cusignal')
            self.xp = importlib.import_module('cupy')
            self.np = importlib.import_module('numpy')
            self.ss = importlib.import_module('scipy.signal')
        else:
            self.xs = importlib.import_module('scipy.signal')
            self.xp = importlib.import_module('numpy')
            self.np = self.xp
            self.ss = self.xs

    def run(self, buff):
        out = self.xp.array(buff)

        if self.dec > 1:
            out = self.xs.resample_poly(out, 1, int(self.dec), window='hamm')

        if self.out_size == -1:
            self.out_size = int(len(buff)/self.dec)

        return self.xs.resample(out, self.out_size, window='hamm')



from setuptools import setup, find_packages
import radio

shortdesc = "Python based library for FM, AM and WBFM demodulation."

setup(
    name='radio-core',
    version=radio.__version__,
    description=shortdesc,
    url='https://github.com/luigifreitas/radio-core',
    packages=find_packages(),
    author='Luigi F. Cruz',
    include_package_data=True
)
